////////////////////////////////////////////////////////////////////////
// \file    Cluster.h
// \brief   Cluster of Hits. Intended for EM showers.
// \version $Id: Cluster.h,v 1.6 2012-11-20 20:26:04 brebel Exp $
// \author  $Author: brebel $
// \date    $Date: 2012-11-20 20:26:04 $
////////////////////////////////////////////////////////////////////////

#ifndef NOVADDT_CLUSTER_H
#define NOVADDT_CLUSTER_H

#include "DDTBaseDataProducts/BaseProducts.h"

namespace novaddt
{
//  class Fragment;

  class Cluster
  {
  public:
    // Default Constructor
    Cluster();

#ifndef __GCCXML__
    
    // Constructor with arguments
    Cluster(const Cell&   lowxc,
	    const Cell&   lowyc,
	    const Cell&   highxc,
	    const Cell&   highyc,
	    const Plane&  lowp,
	    const Plane&  highp,
	    const TDC&    lowt,
	    const TDC&    hight,
	    const uint32_t&    adc);
    
    // Accessor Functions
    Cell	 const& LowCellX()    const;
    Cell	 const& LowCellY()    const;
    Cell	 const& HighCellX()   const;
    Cell	 const& HighCellY()   const;
    Plane	 const& LowPlane()    const;
    Plane	 const& HighPlane()   const;
    TDC		 const& EarliestTDC() const;
    TDC		 const& LatestTDC()   const;
    uint32_t 	 const& TotalADC()    const;
    int  	 const& WidthX()      const;
    int  	 const& WidthY()      const;
    int  	 const& Extent()      const;
    float               Volume()      const;
    
    // Modifier Functions
    void SetLowCellX   (const Cell& lowxc);
    void SetLowCellY   (const Cell& lowyc);
    void SetHighCellX  (const Cell& highxc);
    void SetHighCellY  (const Cell& highyc);
    void SetLowPlane   (const Plane& lowp);
    void SetHighPlane  (const Plane& highp);
    void SetEarliestTDC(const TDC& lowt);
    void SetLatestTDC  (const TDC& hight);
    void SetTotalADC   (const uint32_t& sum); 	   
    void SetWidthX     (const int& widthx);    
    void SetWidthY     (const int& widthy);    
    void SetExtent     (const int& extent);   

#endif

  private:
    Cell	_lowxc;  // Low Cell X
    Cell	_lowyc;  // Low Cell Y
    Cell	_highxc; // High Cell X
    Cell	_highyc; // High Cell Y
    Plane	_lowp;	 // Low Plane
    Plane	_highp;	 // High Plane
    TDC		_lowt;	 // Low Time
    TDC		_hight;	 // High Time
    uint32_t    _adcsum; // Sum of ADC values of hits in Cluster
    int		_widthx;  // Size in x
    int		_widthy;  // Size in y
    int		_extent; // Extent in Z
  };

  inline
    Cluster::Cluster():
    _lowxc  (0),
    _lowyc  (0),
    _highxc (0),
    _highyc (0),
    _lowp   (0),
    _highp  (0),
    _lowt   (0),
    _hight  (0),
    _adcsum (0),
    _widthx (0),
    _widthy (0),
    _extent (0)
      {
      }

#ifndef __GCCXML__
  inline
    Cluster::Cluster(const Cell& lowxc, const Cell& lowyc, const Cell& highxc, const Cell& highyc, const Plane& lowp, const Plane& highp,
		     const TDC& lowt, const TDC& hight, const uint32_t& adc):
    _lowxc  (lowxc),
    _lowyc  (lowyc),
    _highxc (highxc),
    _highyc (highyc),
    _lowp   (lowp),
    _highp  (highp),
    _lowt   (lowt),
    _hight  (hight),
    _adcsum (adc)

    {
    _widthx = highxc - lowxc;
    _widthy = highyc - lowyc;
    _extent = highp  - lowp;
    }

  inline Cell	  const& Cluster::LowCellX()    const { return _lowxc; }
  inline Cell	  const& Cluster::LowCellY()    const { return _lowyc; }
  inline Cell	  const& Cluster::HighCellX()   const { return _highxc; }
  inline Cell	  const& Cluster::HighCellY()   const { return _highyc; }
  inline Plane	  const& Cluster::LowPlane()    const { return _lowp; }
  inline Plane	  const& Cluster::HighPlane()   const { return _highp; }
  inline TDC	  const& Cluster::EarliestTDC() const { return _lowt; }
  inline TDC	  const& Cluster::LatestTDC()   const { return _hight; }
  inline uint32_t const& Cluster::TotalADC()    const { return _adcsum; }
  inline int  	  const& Cluster::WidthX()      const { return _widthx; }
  inline int  	  const& Cluster::WidthY()      const { return _widthy; }
  inline int  	  const& Cluster::Extent()      const { return _extent; }
  inline float           Cluster::Volume()      const { return (_widthx * _widthy * _extent); }
    
  inline void Cluster::SetLowCellX   (const Cell& lowxc)   { _lowxc  = lowxc;  }
  inline void Cluster::SetLowCellY   (const Cell& lowyc)   { _lowyc  = lowyc;  }
  inline void Cluster::SetHighCellX  (const Cell& highxc)  { _highxc = highxc; }
  inline void Cluster::SetHighCellY  (const Cell& highyc)  { _highyc = highyc; }
  inline void Cluster::SetLowPlane   (const Plane& lowp)   { _lowp   = lowp;   }
  inline void Cluster::SetHighPlane  (const Plane& highp)  { _highp  = highp;  }
  inline void Cluster::SetEarliestTDC(const TDC& lowt)     { _lowt   = lowt;   }
  inline void Cluster::SetLatestTDC  (const TDC& hight)    { _hight  = hight;  }
  inline void Cluster::SetTotalADC   (const uint32_t& sum) { _adcsum = sum;    }
  inline void Cluster::SetWidthX     (const int& widthx)   { _widthx = widthx; }
  inline void Cluster::SetWidthY     (const int& widthy)   { _widthy = widthy; }
  inline void Cluster::SetExtent     (const int& extent)   { _extent = extent; }

#endif
  
}

#endif // NOVADDT_CLUSTER_H
/////////////////////////////////////////////////////////////////////////////////////////////
