#ifndef DDT_GROUPED_HIT_LIST_H
#define DDT_GROUPED_HIT_LIST_H


#include "DDTBaseDataProducts/HitList.h"

#include <vector>


namespace novaddt
{
  template <class VAL>
  class GroupedHitList
  {
  public:
    GroupedHitList();

#ifndef __GCCXML__
    typedef VAL value_t;

    // The following link shows how to properly design GroupedHitLists:
    // https://cdcvs.fnal.gov/redmine/projects/novaddt/wiki/Designing_Proper_GroupedHitLists
    template <class It> 
    GroupedHitList(It begin, It end, std::vector<size_t> boundary_list);

    unsigned n_groups() const;

    // this method provides the number of hits in the requested group
    std::size_t group_size(unsigned group) const;

    typename VAL::value_type first_value() const;
    typename VAL::value_type last_value() const;
    typename VAL::value_type extent() const;
    
    // get a sorted hit list for the requested group or grab the entire thing
    HitList get(unsigned group) const;
    HitList const & get() const;
#endif

  private:
#ifndef __GCCXML__
    // get the low and high bound of the requested group
    std::size_t low(unsigned group) const;
    std::size_t high(unsigned group) const;

    // convert the index to an iterator in the ordered hit list
    HitList::const_iterator get_it(size_t index) const;

    // check whether HitList is empty (asserts when it is empty)
    void check_empty() const;

#endif

    HitList hits_;
    std::vector<size_t> bound_list_;
  };
  

  template <class VAL>
  inline GroupedHitList<VAL>::GroupedHitList()
  {
  }


#ifndef __GCCXML__
  

  template <class VAL>
  template <class It> inline 
  GroupedHitList<VAL>::
  GroupedHitList(It begin, It end, std::vector<size_t> boundary_list)
    : hits_(begin, end),
      bound_list_(boundary_list)
  {
  }
    
  
  template <class VAL> inline
  unsigned GroupedHitList<VAL>::n_groups() const
  {
    unsigned result = 0;
    if (!bound_list_.empty())
      // Since the boundary list contains the beginning and the end,
      // the number of groups is one less than the number of boundaries.
      result = bound_list_.size() - 1;

    return result;
  }


  template <class VAL> inline
  std::size_t GroupedHitList<VAL>::low(unsigned group) const
  {
    return bound_list_.at(group);
  }


  template <class VAL> inline
  std::size_t GroupedHitList<VAL>::high(unsigned group) const
  {
    // the upper bound is the lower bound of the next group
    return bound_list_.at(group + 1);
  }


  template <class VAL> inline
  std::size_t GroupedHitList<VAL>::group_size(unsigned group) const
  {
    return high(group) - low(group);
  }


  template <class VAL> inline
  void GroupedHitList<VAL>::check_empty() const
  {
    if (hits_.empty())
    {
      std::cerr << "The HitList is empty, so you cannot retrieve any elements."
		<< std::endl;
      assert(false);
    }
  }


  template <class VAL> inline
  typename VAL::value_type GroupedHitList<VAL>::first_value() const
  {
    check_empty();

    return VAL(hits_.front()).val;
  }


  template <class VAL> inline
  typename VAL::value_type GroupedHitList<VAL>::last_value() const
  {
    check_empty();

    return VAL(hits_.back()).val;
  }


  template <class VAL> inline
  typename VAL::value_type GroupedHitList<VAL>::extent() const
  {
    return last_value() - first_value();
  }


  template <class VAL> inline
  HitList::const_iterator 
  GroupedHitList<VAL>::get_it(std::size_t index) const
  {
    return hits_.begin() + index;
  }


  template <class VAL> inline
  HitList GroupedHitList<VAL>::get(unsigned group) const
  {
    return HitList(get_it(low(group)), get_it(high(group)));
  }


  template <class VAL> inline
  HitList const & GroupedHitList<VAL>::get() const
  {
    return hits_;
  }


#endif


}


#endif
