////////////////////////////////////////////////////////////////////////
// Class:       TriCellTriggerAna
// Module Type: analyzer
// File:        TriCellTriggerAna_module.cc
//
// Generated at Fri Mar  8 16:40:58 2013 by Brian_Rebel using artmod
// from art v1_02_06.
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Core/FindOne.h"

#include "DDTBaseDataProducts/BaseProducts.h"
#include "DDTBaseDataProducts/HitList.h"
#include "DDTBaseDataProducts/DAQHit.h"
#include "DDTBaseDataProducts/Track.h"
#include "DDTBaseDataProducts/TriggerDecision.h"

#include "TH1.h"
#include "TH2.h"

namespace novaddt {
  class TriCellTriggerAna;
}

class novaddt::TriCellTriggerAna : public art::EDAnalyzer {
public:
  explicit TriCellTriggerAna(fhicl::ParameterSet const & p);
  virtual ~TriCellTriggerAna();

  void analyze(art::Event const & e) override;

  void beginJob() override;
  void reconfigure(fhicl::ParameterSet const & p) override;

private:

  float FractionGoodCells(novaddt::HitList const& hl,
			  TH2F*                   occupancy,
			  TH2F*                   occupancyTriCell);

  std::string _triggerDecisionModuleLabel;   ///< label of module making the TD
  std::string _trackModuleLabel;             ///< label of module making the HitList
  std::string _trackInstanceLabel;           ///< instance label making the HitList
  float       _minFraction;                  ///< minimum fraction of tri-cell groups to be used
  TH1F*       _fractionGoodCells;            ///< Fraction of cells in tri-cell groupings
  TH2F*       _cellPlaneOccupancy;           ///< Cell and planes that are passed by the trigger
  TH2F*       _cellPlaneOccupancyTriCell;    ///< Cell and planes that are part of tri-cell groups
  TH2F*       _cellPlaneOccupancyAll;        ///< Cell and planes that are passed by the trigger
  TH2F*       _cellPlaneOccupancyAllTriCell; ///< Cell and planes that are part of tri-cell groups
  TH1F*       _trackLength;                  ///< length of tracks in number of planes
  TH1F*       _tracksPerEvent;               ///< raw number of tracks found in each event
  TH1F*       _goodTracksPerEvent;           ///< tracks passing tricell criteria
  TH1F*       _triggeredTracksPerEvent;      ///< tracks passing tricell trigger in each event
  TH1F*       _triggersPerSecond;            ///< trigger rate in Hz
};

//------------------------------------------------------------------------------
novaddt::TriCellTriggerAna::TriCellTriggerAna(fhicl::ParameterSet const & p)
  : EDAnalyzer(p)
{
  this->reconfigure(p);
}

//------------------------------------------------------------------------------
novaddt::TriCellTriggerAna::~TriCellTriggerAna()
{
}

//------------------------------------------------------------------------------
void novaddt::TriCellTriggerAna::analyze(art::Event const & e)
{
  // get the trigger decisions
  art::Handle<std::vector<novaddt::TriggerDecision>> tdh;
  e.getByLabel(_triggerDecisionModuleLabel, tdh);

  // get the tracks for this event
  art::Handle<std::vector<novaddt::Track> > th;
  e.getByLabel(_trackModuleLabel, _trackInstanceLabel, th);

  // fill histograms of the length of tracks and the number
  // of tracks
  int goodTracks      = 0.;
  _tracksPerEvent->Fill(th->size());

  // get the hit list for each track
  art::FindOne<novaddt::HitList> fohltr(th, e, _trackModuleLabel);
  for(size_t t = 0; t < th->size(); ++t){
    float z0 = th->at(t).StartZ();
    float z1 = th->at(t).EndZ();

    const novaddt::HitList& hl = fohltr.at(t).ref();
    float goodFrac = this->FractionGoodCells(hl, 
					     _cellPlaneOccupancyAll, 
					     _cellPlaneOccupancyAllTriCell);

    _trackLength->Fill(std::abs(z1-z0));
    _fractionGoodCells->Fill(goodFrac);
    if(goodFrac > _minFraction) ++goodTracks;
  }
  _goodTracksPerEvent->Fill(goodTracks);

  // loop over all the trigger decisions and get the associated 
  // hit lists
  art::FindOne<novaddt::HitList> fohl(tdh, e, _triggerDecisionModuleLabel);

  uint32_t triggeredTracks = 0.;
  
  for(size_t t = 0; t < tdh->size(); ++t){
    
    const novaddt::HitList& hl = fohl.at(t).ref();

    float goodFrac = this->FractionGoodCells(hl, _cellPlaneOccupancy, _cellPlaneOccupancyTriCell);

    if(goodFrac > _minFraction) ++triggeredTracks;

    LOG_DEBUG("TriCellTriggerAna") << tdh->at(t).start() 
				   << " " << tdh->at(t).duration()
				   << " " << tdh->at(t).mask()
				   << " " << hl.size();

  }// end loop over trigger decisions

  if(triggeredTracks != tdh->size())
    mf::LogWarning("TriCellTriggerAna") << "Number of trigger decisions is not the same as"
					<< " number of triggered tracks: "
					<< tdh->size() << "/" << triggeredTracks;

  if(triggeredTracks > 0){
    _triggeredTracksPerEvent->Fill(triggeredTracks);
    _triggersPerSecond->Fill(triggeredTracks*1./0.005);
  }

  return;

}

//------------------------------------------------------------------------------
void novaddt::TriCellTriggerAna::beginJob()
{
  art::ServiceHandle<art::TFileService> tfs;

  _fractionGoodCells            = tfs->make<TH1F>("fractionGoodCells", ";Tri-Cell Fraction;Tracks", 
						  100, 0., 1.);       
  _cellPlaneOccupancy           = tfs->make<TH2F>("cellPlaneOccupancy", ";Plane;Cell", 
						  1000, 0., 1000., 512, 0., 512.);
  _cellPlaneOccupancyTriCell    = tfs->make<TH2F>("cellPlaneOccupancyTriCell", ";Plane;Cell", 
						  1000, 0., 1000., 512, 0., 512.);
  _cellPlaneOccupancyAll        = tfs->make<TH2F>("cellPlaneOccupancyAll", ";Plane;Cell", 
						  1000, 0., 1000., 512, 0., 512.);
  _cellPlaneOccupancyAllTriCell = tfs->make<TH2F>("cellPlaneOccupancyAllTriCell", ";Plane;Cell", 
						  1000, 0., 1000., 512, 0., 512.);
  _trackLength                  = tfs->make<TH1F>("trackLength", ";Track Length;Tracks",
						  1000, 0., 1000.);
  _tracksPerEvent               = tfs->make<TH1F>("tracksPerEvent",          ";Tracks;Events", 100, 0, 100);
  _goodTracksPerEvent           = tfs->make<TH1F>("goodTracksPerEvent",      ";Tracks;Events", 100, 0, 100);
  _triggeredTracksPerEvent      = tfs->make<TH1F>("triggeredTracksPerEvent", ";Tracks;Events", 100, 0, 100);
  _triggersPerSecond            = tfs->make<TH1F>("triggersPerSecond",       ";Trigger Rate (Hz); Events", 25, 0, 5000);
  return;
}

//------------------------------------------------------------------------------
void novaddt::TriCellTriggerAna::reconfigure(fhicl::ParameterSet const & p)
{
  _triggerDecisionModuleLabel = p.get<std::string>("TriggerDecisionModuleLabel");
  _trackModuleLabel           = p.get<std::string>("TrackModuleLabel"          );
  _trackInstanceLabel         = p.get<std::string>("TrackInstanceLabel", ""    );
  _minFraction                = p.get<float      >("MinFractionTriCell", 0.6   );
  return;
}

//------------------------------------------------------------------------------
float novaddt::TriCellTriggerAna::FractionGoodCells(novaddt::HitList const& hl, 
						    TH2F*                   occupancy,
						    TH2F*                   occupancyTriCell)
{
  float goodFrac = 0.;

  std::map<unsigned short, std::set<unsigned short>> planesAndCells;

  for(auto const& i : hl){
    planesAndCells[i.Plane().val].insert(i.Cell().val);
    occupancy->Fill(i.Plane().val, i.Cell().val);
  }
  
  // now loop over the planes and cells to determine
  // the fraction in this list that have both neighbors also hit
  float totalCells = 0.;
  float goodCells  = 0.;
  
  for(auto const& pcitr : planesAndCells){
    // get the set for the current plane
    // sets are presorted, go c++
    std::set<unsigned short> cellSet = pcitr.second;
    
    // put the iterator over the set to the second entry
    auto citr = cellSet.cbegin();
    totalCells += 1.;
    ++citr;
    
    // loop over the cells and look to see if there is 
    // a cell in this plane that has both its neighbors also
    // hit on this track
    while( citr != cellSet.cend() ){
      if( cellSet.find(*citr - 1) != cellSet.end() &&
	  cellSet.find(*citr + 1) != cellSet.end() ){
	
	goodCells += 1.;
	occupancyTriCell->Fill(pcitr.first, *citr);
      }

      totalCells += 1.;
      ++citr;
    }// end iterator over cells for this plane
    
  }// end loop to count the tri-cell groupings
  
  // fill the histograms for good tracks, cells
  if(totalCells > 0.) goodFrac = goodCells/totalCells;

  return goodFrac;
}


DEFINE_ART_MODULE(novaddt::TriCellTriggerAna)
