#ifndef DAQHIT_H
#define DAQHIT_H

#include <vector>
#include <memory>   	       

#include "boost/shared_ptr.hpp"

#include "DAQDataFormats/RawEvent.h"
#include "DAQDataFormats/RawDataBlock.h"
#include "DAQDataFormats/RawMicroBlock.h"
#include "DAQDataFormats/RawMilliSlice.h"

#include "NovaDAQConventions/DAQConventions.h"
#include "DAQChannelMap/DAQChannelMap.h"

#include "DDTBaseDataProducts/BaseProducts.h"

/********************************************************************
 * The DAQHit class represents a single unpacked hit in the nova    *
 * detector.  The DAQ Hit has one actual memeber which is a pointer *
 * to the actual data structure that we are going to use for the    *
 * detector readout.  This allows us to transparently change the    *
 * the underlying representation of the data without having to      *
 * mess with the class itself, and to perform shallow copies of the *
 * hits when this is required.                                      *
 * anorman@fnal.gov                                                 *
 ********************************************************************/

namespace novaddt {
  class DAQHitRdr;
  class DAQHitWtr;
  class DAQHit;
}

class novaddt::DAQHit {
public:
  DAQHit();

#ifndef __GCCXML__

  DAQHit(daqdataformats::RawMilliSlice* theMilli,
	 daqdataformats::RawNanoSlice* theNano);

  // Constructor for testing
  DAQHit(novaddt::TDC tdc, novaddt::ADC adc, novaddt::Plane plane);
  // Constructor for RawDigit2DAQHit
  DAQHit(int detID,
	 //novadaq::cnv::DetID  detID,
         daqchannelmap::dchan channel,
         novaddt::ADC         adc,
         novaddt::TDC         tdc
	 );

  daqchannelmap::dchan const& Chan() const {return _theChannel;}

  void reset();

  void print() const;

#endif

  // NOTE:  If the methods below are hidden from ROOT using
  // #ifndef __GCCXML__ then CompareDAQHit fails when trying
  // to make the dictionary.

  novaddt::Plane const& Plane() const { return _thePlane; }
  novaddt::Cell  const& Cell()  const { return _theCell;  }
  novaddt::View  const& View()  const { return _theView;  }
  novaddt::ADC   const& ADC()   const { return _theADC;   }
  novaddt::TDC   const& TDC()   const { return _theTDC;   }

  novaddt::DCM getDCM() const;
  bool FEBOverflowError() const { return _FEBOverflowError; }
  
  // Accessors for the elements based on type (i.e. conversion operators)
  operator novaddt::Plane() const { return Plane();  }
  operator novaddt::Cell()  const { return Cell();   }
  operator novaddt::View()  const { return View();   }
  operator novaddt::ADC()   const { return ADC();    }
  operator novaddt::TDC()   const { return TDC();    }
  operator novaddt::DCM()   const { return getDCM(); } 
  
  bool operator==(const DAQHit& rhs) const
  { return this->Plane() == rhs.Plane() && this->Cell() == rhs.Cell() && this->TDC() == rhs.TDC();}

private:
  novaddt::ADC ADC_Extraction(daqdataformats::RawNanoSlice* theNano) const;

  int                  _theDetID;
  daqchannelmap::dchan _theChannel;
  novaddt::Plane       _thePlane;
  novaddt::Cell        _theCell;
  novaddt::View        _theView;
  novaddt::ADC         _theADC;
  novaddt::TDC         _theTDC;

  bool _FEBOverflowError;
};


#endif
