////////////////////////////////////////////////////////////////////////
// Class:       NovaFragmentFileReader
// Module Type: generator
// File:        NovaFragmentFileReader_module.cc
//
// Generated at Thu Jan 10 09:47:22 2013 by Martin Frank using artmod
// from art v1_02_06.
////////////////////////////////////////////////////////////////////////


#include "artdaq-core/Generators/FragmentGenerator.hh"
#include "artdaq-core/Generators/GeneratorMacros.hh"
#include "art/Utilities/Exception.h"
#include "messagefacility/MessageLogger/MessageLogger.h" // 
#include <string>
#include <vector>
#include <sys/types.h>		// open
#include <sys/stat.h>		// open fstat, struct stat
#include <fcntl.h>		// open
#include "DAQDataFormats/RawMilliBlockHeader.h"
#include "DAQDataFormats/RawMilliBlock.h"
#include "DAQDataFormats/RawMilliSlice.h"
#include "DAQDataFormats/RawMicroSlice.h"


using namespace daqdataformats;


namespace novaddt {
  class NovaFragmentFileReader;
}


class novaddt::NovaFragmentFileReader : public artdaq::FragmentGenerator {
public:
  explicit NovaFragmentFileReader(fhicl::ParameterSet const & p);
  virtual ~NovaFragmentFileReader() noexcept (true);

  bool getNext(artdaq::FragmentPtrs & output) override;


private:
  std::vector<artdaq::Fragment::fragment_id_t> fragmentIDs() override;

  std::vector<int> file_desc;	// vector of file descriptors (fd's)
  std::vector<RawMilliSlice*>	milli_pv; // millislice ptr vector
  std::vector<int> file_sizes;

  unsigned n_reads_;
  artdaq::Fragment::sequence_id_t sequenceID_;

  std::string file_directory_, file_prefix_;
};



novaddt::NovaFragmentFileReader::
NovaFragmentFileReader(fhicl::ParameterSet const & p) : 
  file_desc(12*15, 0), file_sizes(12*15,0), n_reads_(0), sequenceID_(0),
  file_directory_(p.get<std::string>("millislice_file_directory")),
  file_prefix_(p.get<std::string>("millislice_file_prefix"))
{
  // open all the files and get their sizes (for use later/below)
  int fd_cntr = 0;
  for (int diblock = 1; diblock < 16; ++diblock)
  {   
    for (int dcm = 1; dcm < 13; ++dcm)
    {
      std::ostringstream file;
      file << file_directory_ << "/" << file_prefix_
	   << "_MilliSlices_Diblock" << diblock
      	   << "_Dcmnum" << dcm << ".raw";

      int fd = open(file.str().c_str(), O_RDONLY);
      if (fd == -1)
	throw art::Exception(art::errors::FileOpenError)
	  << "Error opening file " << file.str() << ".\n";

      // Place file descriptor into file array:
      file_desc[fd_cntr] = fd;

      struct stat stat_buf;
      int sts = fstat( fd, &stat_buf );
      if (sts == -1)
	throw art::Exception(art::errors::FileOpenError)
	  << "Error stat-ing file " << file.str() << ".\n";
      file_sizes[fd_cntr++] = stat_buf.st_size;

      milli_pv.push_back(new RawMilliSlice
			 (rawmillihdr::MILLISLICEHEADER_SIZE * 2 + 1));
    }
  }

  std::cout << fd_cntr << " millislice files were read in from:" 
	    << "\n\t" << file_directory_ << "/\n" << std::endl;
}



novaddt::NovaFragmentFileReader::~NovaFragmentFileReader() noexcept (true)
{
  for (unsigned ii = 0; ii < file_desc.size(); ++ii)
  {
    int fd = file_desc[ii];
    close(fd);
  }
}



bool novaddt::NovaFragmentFileReader::getNext(artdaq::FragmentPtrs & frags)
{
  ++n_reads_;

  artdaq::FragmentPtrs::size_type incoming_size = frags.size();
  assert(incoming_size == 0);

  // read in slices from files using readData method
  ssize_t slices_bytes = 0;
  for (unsigned ii = 0; ii < file_desc.size(); ++ii)
  {
    int fd = file_desc[ii];

    /*We need to clear the millislice read bytes counter so that the
      readData() method knows to read in a millislice after the
      previous millislice. */
    milli_pv[ii]->softclear();

    /*Denis explained the following to me:
      His MockDataDAQ inserts the millislice version number at the beginning
      of the millislice file, so we need to apply the 4 byte offset when 
      we use lseek below to get to the beginning of the first header.
      For successive millislices, we do not need the offset. */
    unsigned seek_position = 0;
    if (n_reads_ == 1)
      seek_position = 4;
    int lseek_out = lseek(fd, seek_position, SEEK_CUR);
    if (lseek_out < file_sizes[ii])
    {
      readDataStatus rdsts;
      while ((rdsts = milli_pv[ii]->readData((int)fd)) == rds_incomplete)
      {}
      
      if (rdsts == rds_error) 
      {
	std::cerr << "An error occured while reading the millislice data"
		  << " from the file." << std::endl;
	return false;
      }

      slices_bytes += milli_pv[ii]->sizeofdata() * sizeof(unsigned);
    } else {   
      return false;
    }
  }

  // then put all slices into the frag (after the MilliBlockHeader)
  frags.emplace_back
    (new artdaq::Fragment(slices_bytes + sizeof(RawMilliBlockHeader)));
  artdaq::Fragment & frag = *frags.back();

  // set the required fragment identifications
  frag.setFragmentID(0); // one fragment per event only
  frag.setSequenceID(++sequenceID_);


  RawMilliBlockHeader *rmbhp = (RawMilliBlockHeader*)&*frag.dataBegin();
  // millislice data is immediately after the header
  uint8_t *millislices_ptr = (uint8_t*)(rmbhp+1);
  rmbhp->total_bytes = slices_bytes + sizeof(RawMilliBlockHeader);
  rmbhp->seqNum = milli_pv[0]->getMilliSliceSEQ();
  rmbhp->timeStart = 0;
  rmbhp->timeEnd = 0;
  rmbhp->flags.incomplete = 0;
  rmbhp->flags.truncated = 0;
  rmbhp->flags.overflowCorruption = 0;
  rmbhp->flags.expected_count = 0;
  rmbhp->millislice_count = 0;

  // now that the sizes are know, spin through and do the copy.
  rmbhp->timeStart = milli_pv[0]->getTimeStart();
  rmbhp->timeEnd = milli_pv[0]->getTimeStart() + milli_pv[0]->getTimeLength();
  for (unsigned ii = 0; ii < file_desc.size(); ++ii)
  {
    memcpy(millislices_ptr,
	   milli_pv[ii]->getBuffer(),
	   milli_pv[ii]->sizeofdata() * sizeof(unsigned));
    millislices_ptr += milli_pv[ii]->sizeofdata() * sizeof(unsigned);


    // set the milliblock header information
    ++rmbhp->millislice_count;

    if (rmbhp->timeStart > milli_pv[ii]->getTimeStart())
      rmbhp->timeStart = milli_pv[ii]->getTimeStart();

    if (rmbhp->timeEnd < 
	milli_pv[ii]->getTimeStart() + milli_pv[ii]->getTimeLength())
      rmbhp->timeEnd = 
	milli_pv[ii]->getTimeStart() + milli_pv[ii]->getTimeLength();
  }

  return true;
}

std::vector<artdaq::Fragment::fragment_id_t> 
novaddt::NovaFragmentFileReader::fragmentIDs()
{
  return { 0 };
}

DEFINE_ARTDAQ_GENERATOR(novaddt::NovaFragmentFileReader)
