////////////////////////////////////////////////////////////////////////
// Class:       DBSlicer
// Module Type: filter
// File:        DBSlicer_module.cc
//
// Generated at Tue May 13 14:24:37 2014 by Jan Zirnstein using artmod
// from cetpkgsupport v1_05_03.
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Utilities/InputTag.h"
#include "fhiclcpp/ParameterSet.h"

#include "DDTBaseDataProducts/DAQHit.h"
#include "DDTBaseDataProducts/HitList.h"

#include <memory>
#include <string>
#include <float.h>
#include <math.h>

namespace novaddt {
  class DBSlicer;
}

class novaddt::DBSlicer : public art::EDFilter {

public:
  explicit DBSlicer(fhicl::ParameterSet const & p);
  virtual ~DBSlicer();

  bool filter(art::Event & e) override;
  

private:

  // FCL parameters
  std::string fHLabel;
  std::string fInstance;
  float fDistPen;
  float fTimePen;
  float fTimeRes;
  float fEps;
  unsigned int fMinPts;
  int fTimeGap;
  double fMergeKnob;

  unsigned int fSliceID;
  std::vector<std::vector< unsigned int> > fNeighbors;
  std::vector<int> fSliceIDs;

  // Impliment DSCAN algorithm
  void expandSlice(unsigned int Point, std::vector<unsigned int>& Neighbors, unsigned int SliceID);
  std::vector<unsigned int> regionQuery(unsigned int Point){ return fNeighbors[Point]; }
  void InitializeLists(const art::Handle<novaddt::HitList>& Hits);
  double NeighborScore(const art::Handle<novaddt::HitList>& Hits, unsigned int i, unsigned int j);

  // Functions to help split and combine slices
  void GetSingleViewSlice(const std::unique_ptr< std::vector<HitList> >& allClusters,
                        std::unique_ptr< std::vector<HitList> >& desiredClusters, std::vector<double>& desiredTMean,
                        std::vector<double>& desiredPMean, std::vector<double>& desiredTRMS,
                        std::vector<double>& desiredPRMS, daqchannelmap::DetView_TYPE desiredView);

  void MergeSlices(const std::unique_ptr< std::vector<HitList> >& XSlices, 
	std::vector<double>& XTMean, std::vector<double>& XPMean, std::vector<double>& XTRMS, 
	std::vector<double>& XPRMS, const std::unique_ptr< std::vector<HitList> >& YSlices,
	std::vector<double>& YTMean, std::vector<double>& YPMean, std::vector<double>& YTRMS, 
	std::vector<double>& YPRMS, std::unique_ptr< std::vector<HitList> >& CombSlices);

};


novaddt::DBSlicer::DBSlicer(fhicl::ParameterSet const & p)
:
fHLabel   (p.get< std::string  >("hits_label")),
fInstance (p.get< std::string  >("instance")),
fDistPen  (p.get< float        >("Distance_Penalty" )),
fTimePen  (p.get< float        >("Timing_Penalty" )),
fTimeRes  (p.get< float        >("Timing_Resolution" )),
fEps      (p.get< float        >("Epsilon" )),
fMinPts   (p.get< unsigned int >("MinPts" )),
fTimeGap  (p.get< int          >("TimeGap")),
fMergeKnob(p.get< double       >("ViewMergeKnob"))
{
  produces<std::vector<HitList>>("DBSCANSlicesComb");
  produces<std::vector<HitList>>("DBSCANSlicesX");
  produces<std::vector<HitList>>("DBSCANSlicesY");
}

novaddt::DBSlicer::~DBSlicer()
{
  // Clean up dynamic memory and other resources here.
}

void novaddt::DBSlicer::expandSlice(unsigned int Point, std::vector<unsigned int>& Neighbors, unsigned int SliceID)
{
  // Impliment expandSlice
  fSliceIDs[Point] = SliceID;
  for(auto it:Neighbors){
    if(fSliceIDs[it] == -5) {
      fSliceIDs[it] = -1;
      std::vector<unsigned int> MoreNeighbors = regionQuery(it);
      if(MoreNeighbors.size() >= fMinPts) // Add the two neighborhoods
	fNeighbors[Point].insert(fNeighbors[Point].end(), MoreNeighbors.begin(), MoreNeighbors.end());
    // Add hit to slice if it doesn't belong to another one already
    if(fSliceIDs[it] < 0) fSliceIDs[it] = SliceID;
    }
  }
}

double novaddt::DBSlicer::NeighborScore(const art::Handle<novaddt::HitList>& Hits, unsigned int i, unsigned int j)
{
  // Ensure no merging of views at this stage 
  if(Hits->at(i).View() != Hits->at(j).View()) return DBL_MAX;

  // Use FHICL parameter to determine how tight to cluster in space, numbers are cell dimensions in cm
  float DistPenPlane= fDistPen/6.2;
  float TimePenalty = fTimePen/fTimeRes;

  //double dC = abs(Hits->at(i).Cell() - Hits->at(j).Cell());
  double dT = fabs(Hits->at(i).TDC() - Hits->at(j).TDC()) * 15.625; // Conversion to TNS, for coarse timing
  double dP = abs(Hits->at(i).Plane() - Hits->at(j).Plane());

  double Z = dP/DistPenPlane;
  double T = dT/TimePenalty;

  return Z*Z+T*T;
}

void novaddt::DBSlicer::InitializeLists(const art::Handle<novaddt::HitList>& Hits)
{
  fNeighbors.clear();
  fSliceIDs.clear();
  unsigned int size = Hits->size();

  // Initialize IDs to unset
  fSliceIDs.resize(size,-5);

  // Find Neighbors
  fNeighbors.resize(size);
  for(unsigned int it=0; it<size; ++it) {
    fNeighbors[it].push_back(it);
    for(unsigned int jt=it+1; jt<size; ++jt) {
      // If the hits are too far removed in time, don't include any more points
      if(Hits->at(jt).TDC() - Hits->at(it).TDC() > (uint64_t) fTimeGap) break;
      double score = NeighborScore(Hits, it, jt);
      if(score <= fEps) {
	fNeighbors[it].push_back(jt);
	fNeighbors[jt].push_back(it);
      } // end if
    } // end second index
  } // end first index
}

void novaddt::DBSlicer::GetSingleViewSlice(const std::unique_ptr< std::vector<HitList> >& allClusters, 
			std::unique_ptr< std::vector<HitList> >& desiredClusters, std::vector<double>& desiredTMean, 
			std::vector<double>& desiredPMean, std::vector<double>& desiredTRMS, 
			std::vector<double>& desiredPRMS, daqchannelmap::DetView_TYPE desiredView)
{
  for(auto it=allClusters->begin(); it!=allClusters->end(); ++it){
    if((*it)[0].View().val != desiredView) continue;
    desiredClusters->push_back(*it);
    double tsum = 0.0;
    double psum = 0.0;
    double tsum2 = 0.0;
    double psum2 = 0.0;
    for(auto jt:*it){
      tsum  += jt.TDC();
      tsum2 += jt.TDC()*jt.TDC();
      psum  += jt.Plane();
      psum2 += jt.Plane()*jt.Plane();
    desiredTMean.push_back(tsum/it->size());
    desiredPMean.push_back(psum/it->size());
    desiredTRMS.push_back(sqrt(tsum2/it->size()));
    desiredPRMS.push_back(sqrt(psum2/it->size()));
    }
  }
}

void novaddt::DBSlicer::MergeSlices(const std::unique_ptr< std::vector<HitList> >& XSlices, 
	std::vector<double>& XTMean, std::vector<double>& XPMean, std::vector<double>& XTRMS, 
	std::vector<double>& XPRMS, const std::unique_ptr< std::vector<HitList> >& YSlices,
	std::vector<double>& YTMean, std::vector<double>& YPMean, std::vector<double>& YTRMS, 
	std::vector<double>& YPRMS, std::unique_ptr< std::vector<HitList> >& CombSlices)
{
  // This function will have to decide which Cluster from the X View to combine with which Cluster from the Y view
  for(unsigned int it = 0; it<XSlices->size(); ++it){
    // Initialize a few things to keep track of, the minimum score and which Slice it corresponds to
    double minscore = DBL_MAX;
    unsigned int YSlice = 0;
    // A vector to mark which YSlices we have already used to combine with XSlices
    std::vector<unsigned int> used(YSlices->size(),0);
    for(unsigned int jt=0; jt<YSlices->size(); ++jt){
      double score = fabs(XTMean[it]-YTMean[jt])/sqrt(XTRMS[it]*XTRMS[it]+YTRMS[jt]*YTRMS[jt]) + 
		     fabs(XPMean[it]-YPMean[jt])/sqrt(XPRMS[it]*XPRMS[it]+YPRMS[jt]*YPRMS[jt]);
      if(score < minscore && used[jt] == 0){
	minscore = score;
	YSlice = jt;
      }
    }
    // If there is a YSlice that matches the XSlice, put them both into the Combslices, and mark the slice as used
    if(minscore < fMergeKnob){
      novaddt::HitList tmphits(XSlices->at(it));
      tmphits.insert(tmphits.end(), YSlices->at(YSlice).begin(), YSlices->at(YSlice).end());
      CombSlices->push_back(tmphits);
      used[YSlice] = 1;
    }
  }
}

bool novaddt::DBSlicer::filter(art::Event & e)
{
  art::Handle<novaddt::HitList> hits;
  e.getByLabel(fHLabel, fInstance, hits);

  if(hits->empty()) return false;
  InitializeLists(hits);

  // Meat and Potatoes of DBSCAN
  fSliceID = 0;
  for(unsigned int it = 0; it < hits->size(); ++it) {
    if(fSliceIDs[it] != -5) continue;
    // Mark hit as "visited"
    fSliceIDs[it] = -1;
    std::vector<unsigned int> Neighbors = regionQuery(it);
    // Bail if the hit doesn't have enough neighbors
    if(Neighbors.size() < fMinPts) fSliceIDs[it] = 0;
    else {
      // Make a new slice and figure out who goes into the slice
      fSliceID++;
      expandSlice(it,Neighbors,fSliceID);
    } // end else
  } //for all unvisited hits

  // Initialize the product
  std::unique_ptr< std::vector<HitList> >product(new std::vector<HitList>);

  // Resize it to the number of HitLists we'll be producing
  product->resize(fSliceID);
  // Fill the product HitLists with the corresponding DAQHits from the original Handle
  for(unsigned int it = 0; it<fSliceIDs.size(); ++it) 
    product->at(fSliceIDs[it]).push_back(hits->at(it));

  // Now use that product to fill the things we are actually going to put in the event, 
  // view split 2D slices and a view combined slicelist of hitlists
  std::unique_ptr< std::vector<HitList> >productcomb(new std::vector<HitList>);
  std::unique_ptr< std::vector<HitList> >productx(new std::vector<HitList>);
  std::unique_ptr< std::vector<HitList> >producty(new std::vector<HitList>);

  // Containers for storing information to be used in the merging step
  std::vector<double> XMeanTime, XMeanPlane, XRMSTime, XRMSPlane;
  std::vector<double> YMeanTime, YMeanPlane, YRMSTime, YRMSPlane;

  GetSingleViewSlice(product, productx, XMeanTime, XMeanPlane, XRMSTime, XRMSPlane, daqchannelmap::X_VIEW);
  GetSingleViewSlice(product, producty, YMeanTime, YMeanPlane, YRMSTime, YRMSPlane, daqchannelmap::Y_VIEW);
  MergeSlices(productx, XMeanTime, XMeanPlane, XRMSTime, XRMSPlane, 
	      producty, YMeanTime, YMeanPlane, YRMSTime, YRMSPlane, productcomb);

  //put the hits into the event for later use
  e.put(std::move(productcomb),"DBSCANSlicesComb");  
  e.put(std::move(productx),"DBSCANSlicesX");  
  e.put(std::move(producty),"DBSCANSlicesY");  
  return true;

}

DEFINE_ART_MODULE(novaddt::DBSlicer)
