////////////////////////////////////////////////////////////////////////
// Class:       FEBFlashFilter
// Module Type: filter
// File:        FEBFlashFilter_module.cc
//
// Generated at Tue Feb  3 09:00:52 2015 by Jan Zirnstein using artmod
// from cetpkgsupport v1_07_01.
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Utilities/InputTag.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include "DDTBaseDataProducts/DAQHit.h"
#include "DDTBaseDataProducts/HitList.h"

#include <memory>
#include <cstdint>

namespace novaddt {
  class FEBFlashFilter;
}

class novaddt::FEBFlashFilter : public art::EDFilter {
public:
  explicit FEBFlashFilter(fhicl::ParameterSet const & p);
  // The destructor generated by the compiler is fine for classes
  // without bare pointers or other resource use.

  // Plugins should not be copied or assigned.
  FEBFlashFilter(FEBFlashFilter const &) = delete;
  FEBFlashFilter(FEBFlashFilter &&) = delete;
  FEBFlashFilter & operator = (FEBFlashFilter const &) = delete;
  FEBFlashFilter & operator = (FEBFlashFilter &&) = delete;

  // Required functions.
  bool filter(art::Event & event) override;


private:

  // Declare member data here.
  std::string _sortedListLabel;
  std::string _sortedListInstance;
  unsigned int _saturatedADC;
  unsigned int _flasherADC;
  unsigned int _FEBDeadTime;
  unsigned int _numFlasherHits;
  std::map<unsigned int, std::vector<uint64_t>> _saturatedFEBMap;

  void _addToMap(const novaddt::DAQHit& Hit);
  bool _isItAFlasher(const novaddt::DAQHit& Hit);
  unsigned int _febNum(const novaddt::DAQHit& Hit);
};


novaddt::FEBFlashFilter::FEBFlashFilter(fhicl::ParameterSet const & p):
// Initialize member data here.
_sortedListLabel    (p.get< std::string >("hits_label")),
_sortedListInstance (p.get< std::string >("instance")),
_saturatedADC    (p.get< unsigned int   >("saturated_ADC")),
_flasherADC      (p.get< unsigned int   >("flasher_ADC")),
_FEBDeadTime     (p.get< unsigned int   >("FEB_dead_time"))
{
  std::cout << "--- novaddt::FEBFlashFilter Runtime Configuration"    << std::endl;
  std::cout << "\t Input hitlists label: "     << _sortedListLabel    << std::endl;
  std::cout << "\t Input hitlists instance: "  << _sortedListInstance << std::endl;
  std::cout << "\t ADC Saturation Threshold: " << _saturatedADC       << std::endl;
  std::cout << "\t Flasher ADC Threshold: "    << _flasherADC         << std::endl;
  std::cout << "\t FEB Masked off time: "      << _FEBDeadTime        << std::endl;

  // Call appropriate produces<>() functions here.
  produces<HitList>("FEBFlashFiltered");
}

bool novaddt::FEBFlashFilter::filter(art::Event & event)
{
  // Implementation of required member function here.
  // Start by reading in the hit list
  art::Handle<novaddt::HitList> hits;
  event.getByLabel(_sortedListLabel, _sortedListInstance, hits);

  // If the hitlist is empty, bail at this point
  if(hits->empty()) return false;

  // initialize number of flashers to be updated
  _numFlasherHits = 0;

  // Make a hitlist out of hits that survive the FEBFlashFilter
  std::unique_ptr<novaddt::HitList>product(new novaddt::HitList);

  // Now do the actual heavy lifting
  // If there is a saturated hit, add the FEB to a list/map
  // Add the time of the hit so we can reject subsequent hits
  for(auto ihit = hits->begin(); ihit < hits->end(); ++ihit){

    unsigned int hitADC = ihit->ADC().val;

    // If it's a saturated hit, add the info to the map
    if(hitADC >= _saturatedADC) {
      _addToMap(*ihit);
//      std::cout << "FEB hit: " << _febNum(*ihit) << " with Charge: " << hitADC << " at time: " << ihit->TDC().val << std::endl;
    }

    // If it's a flasher, don't add it to the product, just continue
    else if(_isItAFlasher(*ihit)) continue;

    // At this point we have saturated hits, and non flashers, so add them to the product
    product->push_back(*ihit);
  }

  // Put the filtered list into the event and we're done
  event.put(std::move(product),"FEBFlashFiltered");

/* Uncomment this block if you'd like some debug output about the Flasher Map

  for(auto it= _saturatedFEBMap.cbegin(); it != _saturatedFEBMap.cend(); ++it){
    std::cout << "Key: " << it->first << std::endl << "Values: " << std::endl;
    for(auto vit = it->second.begin(); vit != it->second.end(); ++vit) {
      std::cout << "  " << *vit << std::endl;
    }
  }

  std::cout << "Removed " << _numFlasherHits <<  " from the event" << std::endl;
*/

  return true;
}

void novaddt::FEBFlashFilter::_addToMap(const novaddt::DAQHit& Hit)
{
  // Need FEB and TDC of hit
  auto hitTDC = Hit.TDC().val;
  //double hitTDC = Hit.TDC().val;
  unsigned int hitFEB = _febNum(Hit);
  // OK will have to compute FEB from Plane and Cell
  // Efficient inserting into maps (with hint position)
  if(_saturatedFEBMap.count(hitFEB) > 0) _saturatedFEBMap.at(hitFEB).push_back(hitTDC);
  else {
    auto it = _saturatedFEBMap.begin();
    std::vector<uint64_t> temp_insert (1,hitTDC);
    _saturatedFEBMap.insert(it, std::pair<unsigned int, std::vector<uint64_t>>(hitFEB,temp_insert));
  }

}

bool novaddt::FEBFlashFilter::_isItAFlasher(const novaddt::DAQHit& Hit)
{
  unsigned int febnum = _febNum(Hit);
  unsigned int hitADC = Hit.ADC().val;

  // If the ADC of the hit are above flasher threshold, it's not a flasher
  if(hitADC >= _flasherADC) return false;

  // if the hit doesn't have it's FEB in the map yet, it's not a flasher
  else if(_saturatedFEBMap.count(febnum) < 1) return false;

  // if the FEB is in the map, have to check timestamps to determine flasher status
  auto hitTDC = Hit.TDC().val;
  uint64_t conv = 15625; //15.625 ns per TDC tick

  for(auto& elem: _saturatedFEBMap.at(febnum)){
    if((hitTDC >= elem) && (hitTDC <= (elem + (conv*_FEBDeadTime)))) {
      _numFlasherHits++;
      return true; 
    }
  }

  // If we got to this point it's not a flasher
  return false;
}

unsigned int novaddt::FEBFlashFilter::_febNum(const novaddt::DAQHit& Hit)
{
  // Let's define the FEB number as:
  // Cell number div 32 plus plane number * 12
  unsigned int febnum;
  unsigned int hitCell  = Hit.Cell().val;
  unsigned int hitPlane = Hit.Plane().val;

  febnum = hitPlane*100 + hitCell/32;

  return febnum;
}

DEFINE_ART_MODULE(novaddt::FEBFlashFilter)
