////////////////////////////////////////////////////////////////////////
// Class:       RemoveSpatialNoise
// Module Type: producer
// File:        RemoveSpatialNoise_module.cc
//
// Generated at Wed June 5 by Matthew Tamsett based on SpaceSlice
// 
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Event.h"

#include "DDTBaseDataProducts/BaseProducts.h"
#include "DDTBaseDataProducts/HitList.h"
#include "DDTBaseDataProducts/DAQHit.h"
#include "DDTBaseDataProducts/CompareDAQHit.h"

#include "DAQChannelMap/DAQChannelMap.h"
//---------------------------------------------------------------------
namespace novaddt {
    class RemoveSpatialNoise;
}
//---------------------------------------------------------------------
class novaddt::RemoveSpatialNoise : public art::EDFilter{
    public:
        explicit RemoveSpatialNoise(fhicl::ParameterSet const & p);
        virtual ~RemoveSpatialNoise();
        virtual bool filter(art::Event & event);
        void endJob() override;
    private:
        // Declare member data here.
        std::string     fInputLabel;
        std::string     fInputInstance;
        int             fMaxXYGap;
        unsigned int    fMinXYHits;
        bool            sorted_by_cell;
        // functions
        void pruneOutlyingCells(HitList & hits,
                                novaddt::View view);
        void printDDTSlice(HitList & hits);
        // Counters
        int _nEvents        = 0;
        int _nInputSlices   = 0;
        int _nOutputSlices  = 0;
        int _nHitsPruned    = 0;
};
//---------------------------------------------------------------------
novaddt::RemoveSpatialNoise::RemoveSpatialNoise(fhicl::ParameterSet const & p)
:
    fInputLabel     (p.get< std::string >("InputLabel"   )),
    fInputInstance  (p.get< std::string >("InputInstance")),
    fMaxXYGap       (p.get< int         >("MaxXYGap"     )),
    fMinXYHits      (p.get< int         >("MinXYHits"    ))
    // Initialize member data here.
{
    std::cout << "=== novaddt::RemoveSpatialNoise instantiate" << std::endl;
    std::cout << "\t input hits label:        " << fInputLabel      << std::endl;
    std::cout << "\t input hits instance:     " << fInputInstance   << std::endl;
    std::cout << "\t Max XY gap:              " << fMaxXYGap        << std::endl;
    std::cout << "\t Min X/Y hits:            " << fMinXYHits       << std::endl;
    // Call appropriate Produces<>() functions here.
    produces<std::vector<HitList>>("CleanTimeSpaceSlice");
}
//---------------------------------------------------------------------
novaddt::RemoveSpatialNoise::~RemoveSpatialNoise()
{
    // Clean up dynamic memory and other resources here.
}
//---------------------------------------------------------------------
bool  novaddt::RemoveSpatialNoise::filter(art::Event & event)
{
    LOG_DEBUG("RemoveSpatialNoise") << "=== novaddt::RemoveSpatialNoise filter. Event: " 
              << event.id().event() 
              << std::endl;
    _nEvents++;
    art::Handle<std::vector<HitList> > hit_lists;
    event.getByLabel(fInputLabel, fInputInstance, hit_lists);
    LOG_DEBUG("RemoveSpatialNoise") << "\t got " << hit_lists->size() << " hit lists" << std::endl;
    
    //make a hit lists out of hits/hit lists that pass the spatial noise removal
    std::unique_ptr< std::vector<HitList> >final_product(new std::vector<HitList>);
    
    for(unsigned int i=0; i<hit_lists->size(); ++i){
        _nInputSlices++;
        HitList product = hit_lists->at(i);
        LOG_DEBUG("RemoveSpatialNoise") << "\t hit list[" << i << "]: " << product.size() << " hits" << std::endl;
        //printDDTSlice(product);
        
        // prune outliers in X
        sorted_by_cell = false;
        pruneOutlyingCells(product, novaddt::View(daqchannelmap::X_VIEW));
        // prune outliers in Y
        sorted_by_cell = true;
        pruneOutlyingCells(product, novaddt::View(daqchannelmap::Y_VIEW));
        
        //printDDTSlice(product);
        if (product.size() > 0) final_product->push_back(product);
    } // end of loop on hits lists
    //put the hits into the event for later use
    LOG_DEBUG("RemoveSpatialNoise") << "\t produced " << final_product->size() << " hit lists" << std::endl;
    _nOutputSlices+=final_product->size();
    event.put(std::move(final_product),"CleanTimeSpaceSlice");

    return true;
}
//---------------------------------------------------------------------
void  novaddt::RemoveSpatialNoise::pruneOutlyingCells(HitList & hits,
                                                      novaddt::View view)
{
    unsigned int n_hits = hits.size();
    
    LOG_DEBUG("RemoveSpatialNoise") << "\t\t\t hit list, has " << n_hits << " hits" << std::endl;
    
    if (!sorted_by_cell) std::sort(hits.begin(), hits.end(), CompareDAQHit<Cell>());
    
    std::vector<int> current_slice_indices;        
    std::vector<int> to_erase;        
    int current_end = -9; // dummy value to start with
    
    for (unsigned int i_hit = 0; i_hit < n_hits; ++i_hit){
        DAQHit hit = hits[i_hit];
        
        // check that the current cell is in the correct view
        if (!(hit.View() == view)) continue;
        //LOG_DEBUG("RemoveSpatialNoise") << "\t\t - found hit in this view: " << i_hit << std::endl;
        // if current end not yet initialised do this now
        if (current_end == -9){
            current_end = i_hit;
            current_slice_indices.push_back(i_hit);
            continue;
        }
        
        const int thisCell = hit.Cell().val;
        const int endCell  = hits[current_end].Cell().val;
        
        //LOG_DEBUG("RemoveSpatialNoise") << "\t\t - xy gap[" << current_end  << "," << i_hit 
                  //<< "]: "            << endCell
                  //<< " - "            << thisCell
                  //<< std::endl;
        
        // check for a large seperation
        if(thisCell - endCell > fMaxXYGap){
            LOG_DEBUG("RemoveSpatialNoise") << "\t\t\t - large gap[" << current_end  << "," << i_hit 
                      << "]: "                 << endCell
                      << " - "                 << thisCell
                      << ", current slice size: " 
                      << current_slice_indices.size() << std::endl;
            // if the current slice is small, mark the hits for removal
            if (current_slice_indices.size() < fMinXYHits){
                LOG_DEBUG("RemoveSpatialNoise") << "\t\t\t - erase small slice: " 
                          << current_slice_indices.size() << " hits"
                          << std::endl;
                for (auto const &i: current_slice_indices){
                    to_erase.push_back(i);
                }
            }
            // restart the current slice
            current_slice_indices.clear();
            current_slice_indices.push_back(i_hit);
            current_end = i_hit;
        } else {
            current_slice_indices.push_back(i_hit);
            current_end = i_hit;
        }
    } // end of loop on hits
    
    LOG_DEBUG("RemoveSpatialNoise") << "\t\t - end of loop, current slice size: " 
                      << current_slice_indices.size() << std::endl;
    // finalise
    if (current_slice_indices.size() < fMinXYHits){
        LOG_DEBUG("RemoveSpatialNoise") << "\t\t\t - erase small final slice: " 
                  << current_slice_indices.size() << " hits"
                  << std::endl;
        for (auto const &i: current_slice_indices){
            to_erase.push_back(i);
        }
    } // end of finalise
    
    // erase outliers
    LOG_DEBUG("RemoveSpatialNoise") << "\t\t\t - erasing " << to_erase.size() << " hits" << std::endl;
    // sort in reverse order
    std::sort(to_erase.begin(), to_erase.end(), std::greater<int>());
    for (auto const &i: to_erase){
        LOG_DEBUG("RemoveSpatialNoise") << "\t\t\t\t - erasing " << i << std::endl;
        hits.erase(hits.begin()+i);
        _nHitsPruned++;
    }
    LOG_DEBUG("RemoveSpatialNoise") << "\t\t\t - final slice size: " << hits.size() << std::endl;
}
//---------------------------------------------------------------------
void novaddt::RemoveSpatialNoise::printDDTSlice(HitList & hits)
{
    // print them
    LOG_DEBUG("RemoveSpatialNoise") << "\t\t - Hit list: " << hits.size()
              << std::endl;
    // sort them
    sort(hits.begin(), hits.end(), novaddt::CompareDAQHit<novaddt::Cell>());
    // loop over hits in each slice
    for(unsigned int j = 0; j < hits.size(); ++j){
        novaddt::DAQHit hit(hits[j]);
        std::string view = "y";
        if( hit.View().val == daqchannelmap::X_VIEW ){
            view = "x";
        }
        LOG_DEBUG("RemoveSpatialNoise") << "\t\t\thit[" << j
                  << "]: TDC: "   << hit.TDC().val
                  << ", ADC: "    << hit.ADC().val
                  << ", plane: "  << hit.Plane().val
                  << ", "         << view
                  << "-cell: "    << hit.Cell().val
                  << std::endl;            
    } // end of loop on hits        
}
//---------------------------------------------------------------------
void novaddt::RemoveSpatialNoise::endJob()
{
    std::cout << "=== novaddt::RemoveSpatialNoise endJob" << std::endl;
    std::cout << "\tNumber of events:           " << _nEvents       << std::endl;
    std::cout << "\tNumber of slices:           " << _nInputSlices  << std::endl;
    std::cout << "\tNumber of slices filtered:  " << _nOutputSlices << std::endl;
    std::cout << "\tNumber of hits pruned:      " << _nHitsPruned   << std::endl;
}
//---------------------------------------------------------------------
DEFINE_ART_MODULE(novaddt::RemoveSpatialNoise)
