//////////////////////////////////////////////////////////////////////
// Class:       SingletonRejection
// Module Type: producer
// File:        SingletonRejection_module.cc
//
// Generated at Tues June 25 2013 by Will Henderson
// Documentation at DOCDB 9475
// Modified by Matthew Tamsett Wed May 28 2014
// This should run after time sorting and before anything else
// if input hits are not time sorted then the output will be rubbish
//////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Event.h"

#include "DDTBaseDataProducts/BaseProducts.h"
#include "DDTBaseDataProducts/HitList.h"
#include "DDTBaseDataProducts/GroupedHitList.h"
#include "DDTBaseDataProducts/DAQHit.h"
#include "DDTBaseDataProducts/CompareDAQHit.h"

#include "DAQChannelMap/DAQChannelMap.h"

//////////////////////////////////////////////////////////////////////
namespace novaddt {
    class SingletonRejection;
}
//////////////////////////////////////////////////////////////////////
class novaddt::SingletonRejection : public art::EDFilter {
public:
    explicit SingletonRejection(fhicl::ParameterSet const & p);
    virtual ~SingletonRejection();
    virtual bool filter(art::Event & event);
    void endJob() override;
private:
    // Declare member data
    std::string     fInputLabel;
    std::string     fInputInstance;
    int             fMinCellSep;
    int             fMinPlaneSep;
    int             fMinTDCSep;
    // Functions
    void filterNoise(art::Handle<novaddt::HitList>  & hits, std::unique_ptr< novaddt::HitList > & product); // Use virtual and pointers for efficiency (unimplemented)
    // Counters
    int _nEvents        = 0;
    int _nHitsIn        = 0;
    int _nHitsOut       = 0;
};
//////////////////////////////////////////////////////////////////////
novaddt::SingletonRejection::SingletonRejection(fhicl::ParameterSet const & p)
:
fInputLabel     (p.get< std::string       >("InputLabel"   )),
fInputInstance  (p.get< std::string       >("InputInstance")),
fMinCellSep     (p.get< int               >("MinCellSep"   )),
fMinPlaneSep    (p.get< int               >("MinPlaneSep"  )),
fMinTDCSep      (p.get< int               >("MinTDCSep"    ))
// Initialize member data
{
    std::cout << "--- novaddt::SingletonRejection begin" << std::endl;
    std::cout << "\t Input hitlists label:    " << fInputLabel    << std::endl;
    std::cout << "\t Input hitlists instance  " << fInputInstance << std::endl;
    std::cout << "\t Min X or Y separation:   " << fMinCellSep    << std::endl;
    std::cout << "\t Min Z separation:        " << fMinPlaneSep   << std::endl;
    std::cout << "\t Min time separation:     " << fMinTDCSep     << std::endl;
    
    // Call appropriate Produces<>() functions
    produces<novaddt::HitList>("SingletonRejected");
}
//////////////////////////////////////////////////////////////////////
novaddt::SingletonRejection::~SingletonRejection()
{
    // Clean up dynamic memory and other resources here.
}
//////////////////////////////////////////////////////////////////////
bool  novaddt::SingletonRejection::filter(art::Event & event)
{
    _nEvents++;
    
    // Array of hitlist from event (input)
    art::Handle<novaddt::HitList> hits;
    event.getByLabel(fInputLabel, fInputInstance, hits);
    _nHitsIn+=hits->size();
    
    // Array of filtered hits (output)
    std::unique_ptr< novaddt::HitList >final_product(new novaddt::HitList);
    
    // Remove noise 
    filterNoise(hits, final_product);
    

    // Add filtered hitlists to event
    _nHitsOut+=final_product->size();
    event.put(std::move(final_product),"SingletonRejected");
    
    return true;
}
//////////////////////////////////////////////////////////////////////
void  novaddt::SingletonRejection::filterNoise(art::Handle<novaddt::HitList> & hits, std::unique_ptr< novaddt::HitList > & product)
{
    unsigned int n_hits = hits->size(); // # of hits in this hitlist
    
    unsigned int lower_limit = 0; // the lower index to start checking for a correlated hit at
    
    for (unsigned int i=0; i<n_hits; ++i){
        //std::cout << "\thit["    << i
                  //<< "], TDC: "  << hits->at(i).TDC().val
                  //<< ", View: "  << hits->at(i).View().val
                  //<< ", Plane: " << hits->at(i).Plane().val
                  //<< ", Cell: "  << hits->at(i).Cell().val
                  //<< std::endl;
        for (unsigned int j=lower_limit; j<n_hits; ++j){
            if (i == j) continue;
            // check if hits are causally correlated
            if (fabs((int)hits->at(i).TDC().val - (int)hits->at(j).TDC().val) > fMinTDCSep){
                // if they aren't & j is lower than i
                // then we ignore this hit 
                // & next time we start the inner loop at this point
                if (j < i){
                    lower_limit = j;
                    continue;
                // similarly if j is the next hit in the list 
                // then j can also not be correlated with i
                // so we advance the lower limit now
                // and break the inner loop
                } else if (j == i+1){
                    lower_limit = j;
                    break;
                // else break inner loop here
                } else {
                    break;
                }
            } // end of check on causal correlation
            // check view correlation
            if (hits->at(i).View().val != hits->at(j).View().val) continue;
            // check plane correlation
            // have to cast these to ints so we can handle negative values
            if (fabs((int)hits->at(i).Plane().val - (int)hits->at(j).Plane().val) > fMinPlaneSep) continue;
            // check cell correlation
            if (fabs((int)hits->at(i).Cell().val  - (int)hits->at(j).Cell().val)  > fMinCellSep)  continue;
            product->push_back(hits->at(i));
            break;
        } // end of inner loop on hits
    } // end of outer loop on hits
    
    
    return;
}
//////////////////////////////////////////////////////////////////////
void novaddt::SingletonRejection::endJob()
{
    std::cout << "--- novaddt::SingletonRejection end " << std::endl;
    std::cout << "\t# of events:        " << _nEvents   << std::endl;
    std::cout << "\t# of hits input:    " << _nHitsIn   << std::endl;
    std::cout << "\t# of hits output:   " << _nHitsOut  << std::endl;
}
//////////////////////////////////////////////////////////////////////
DEFINE_ART_MODULE(novaddt::SingletonRejection)
